`include "slave_ahb.v"

module master(  
		//global inputs
		HCLK,HRESETn,
		//inputs
		HGRANT,BUSREQ,ADDREQ, //arbitration signals - for multi-master

		WRITE, //WRITE: Write request from the processor
		ADDR,// Requested address by the processor (provided by the testbench)
		WDATA,//Write data from the processor (provided by the testbench)
		SIZE,BURST,
		SEL,TRANS,

		HREADY,
		HRESP,
		HRDATA,

		//outputs
		HBUSREQ,HLOCK,hcount,HSIZE,
		
		HSEL,
		HADDR,
		HBURST,
		HTRANS,
		HWDATA,
		HWRITE,
               hready_i //added
		); 

//slave

slave_ahb slave_1( 
               .hclk          (HCLK	), 
               .hresetn       (HRESETn	), 

               .hsel_i        (HSEL	), 
               .haddr_i       (HADDR	), //   .haddr_s       ({HADDR,2'b00} ), 
               .hburst_i      (HBURST	), 
               .htrans_i      (HTRANS	), 
               .hwdata_i      (HWDATA	), 
               .hwrite_i     (HWRITE	), 
		.hsize_i(SIZE),
               .hrdata_o      (HRDATA ), ///
               .hready_o      (HREADY	), 
               .hresp_o      (HRESP	),
                .hready_i ( hready_i)
                 );

//added
output hready_i ;





input HCLK,HRESETn;
input HGRANT,HREADY,BUSREQ,ADDREQ,WRITE; 
input [31:0]WDATA; 
input [3:0]ADDR; 

input [2:0]SIZE,BURST; 
input [1:0]HRESP,TRANS; 
input SEL; 
input [31:0]HRDATA;/////////

output HBUSREQ,HLOCK,HWRITE ,hcount; 
output [1:0]HTRANS;
output HSEL; 

output [31:0]HWDATA; 
output [3:0]HADDR; 
output [2:0]HSIZE,HBURST; 
 
wire HCLK,HRESETn;

wire HGRANT,HREADY,WRITE; 
wire [31:0]WDATA; 
wire [3:0]ADDR;
wire [2:0]SIZE,BURST; 
wire [1:0]HRESP,TRANS; 
wire SEL; 
wire [31:0]HRDATA; 
reg  hready_i = 1'b1;

reg HBUSREQ,HLOCK,HWRITE,hcount; 
reg [1:0]HTRANS; 
reg HSEL ; 

reg [31:0]HWDATA; 
reg [3:0]HADDR;
reg [2:0]HSIZE,HBURST; 
 
reg bus_reg,addr_reg,new_hready,old_hready; 
reg [31:0] RDATA; 
reg [3:0] h_addr; 
 
parameter OKAY = 2'b00, 
          ERROR = 2'b01, 
          RETRY = 2'b10, 
          SPLIT = 2'b11; 
assign HRDATA  = 1;       ////////////////////////////  
always @(posedge HCLK) 
begin 
   if(!HRESETn)              //Master reset 
   begin 
      HBUSREQ = 0; 
      HLOCK = 0; 
      HWRITE = 0; 
      HTRANS = 2'b00; 
      HSEL = 1'b0;
      HADDR = 4'h0; 
      HWDATA = 32'h00000000; 
      HSIZE = 2'b00; 
      HBURST = 2'b00; 
      bus_reg = 0; 
      addr_reg = 0; 
      new_hready = 0; 
      old_hready = 0; 
      hcount = 0; 
   end 
end 
 
always @(posedge HCLK)       //Master sending the request signal to the arbiter 
begin 
   if(!bus_reg) 
   begin 
      if(BUSREQ) 
      begin 
        HBUSREQ = 1'b1; 
        bus_reg = 1; 
      end 
     else if(!BUSREQ) 
        HLOCK = 1'b0;  
   end 
    
   else if(bus_reg) 
      begin 
      HBUSREQ = 1'b0; 
      bus_reg = 0; 
      if(HGRANT) 
        HLOCK = 1'b1; 
      end 
    
end 
 
always@(posedge HCLK)      
begin  
  if(HRESETn) 
  begin  
    if(HGRANT) 
    begin 
      if(!addr_reg) 
      begin 
         if(ADDREQ)        //Master sending the address and the control signals once the bus is granted 
         begin 
          HADDR = ADDR; 
          h_addr = ADDR; 
          HWRITE = WRITE; ////////////
          HSIZE = SIZE; 
          HBURST = BURST; 
          HSEL = SEL; 
          HTRANS = TRANS; 
          addr_reg = 1'b1; 
          HWDATA = 32'h00000000; 
         end 
      end 
     else if(addr_reg) //////// zeros come from here
      begin 
          HADDR = 4'h0; 
          HWRITE = 1'b0; 
          HSIZE = 3'b000; 
          HBURST = 3'b000; 
          HTRANS = 2'b00; 
          addr_reg = 1'b0; 
      end 
     
     
    if(!ADDREQ)              // DATA TRANSFER 
    begin 
      if(WRITE) 
      begin 
            hcount = 0; 
        case ({TRANS}) 
         
             2'b00 : begin  //NON SEQUENTIAL  
                     HWDATA = WDATA ; /// here the data modification should be 
                     if(HREADY && !new_hready && HRESP == OKAY) 
                         new_hready = 1; 
                     else if (new_hready != old_hready) 
                         HWDATA = 32'h00000000; 
                     end 
             2'b01 : begin  // SEQUENTIAL  
                     hcount = hcount + 1; 
                     new_hready = 0; 
                     HWDATA = WDATA ; 
                     if(HREADY && !new_hready && HRESP == ERROR) 
                         new_hready = 1; 
                     else if (new_hready != old_hready) 
                         HWDATA = 32'h00000000; 
                     end                      
             2'b10 : begin  // IDLE  
                     HWDATA = 32'h00000000; 
                     end  
             2'b11 : begin // BUSY 
                     hcount = hcount + 1; 
                     HWDATA = WDATA ; 
                     if(HREADY && HRESP == OKAY) 
                     begin 
                         if(!new_hready) 
                            new_hready = 1;                          
                     end     
                     else if (new_hready != old_hready) 
                     begin  
                         HWDATA = WDATA; 
                         new_hready = 0; 
                     end 
                     else if (HREADY && HRESP == ERROR) 
                     begin 
                        HWDATA = 32'h00000000; 
                     end 
                     end 
      endcase                   
    end 
      else if(!WRITE) 
      begin 
        case ({TRANS}) 
         
             2'b00 : begin  //NON SEQUENTIAL  
                     if(!HREADY) 
                       RDATA = HRDATA; 
                     else if(HREADY) 
                       RDATA = 32'h00000000;                          
                     end 
             2'b01 : begin  // SEQUENTIAL  
                     if(!HREADY) 
                     begin 
                        RDATA = HRDATA; 
                        if(HBURST == 000) 
                           h_addr = h_addr + 1; 
                        else 
                           h_addr = h_addr - 1;  
                     end 
                     else if(HREADY) 
                       RDATA = 32'h00000000;        
                     end                      
             2'b10 : begin  // IDLE  
                     RDATA = 32'h00000000;   
                     end  
             2'b11 : begin // BUSY 
                     if(!HREADY) 
                     begin 
                        RDATA = HRDATA; 
                        if(HBURST == 000) 
                           h_addr = h_addr + 1; 
                        else 
                           h_addr = h_addr - 1;  
                     end       
                     else if(HREADY) 
                         RDATA = HRDATA;  
                     end 
      endcase                   
    end 
    
    
   end        
   end  
     
  end 
  end 
 
endmodule
 
